import datetime
from typing import FrozenSet, Iterable

from django.utils.http import parse_http_date
from rest_framework.request import Request

from .errors import BadRequest

__all__ = (
    "HeaderError",
    "MissingHeader",
    "InvalidHeader",
    "parse_header_datetime",
    "parse_header_etags",
)


class HeaderError(ValueError):
    code = "invalid_header"

    def as_api_error(self) -> BadRequest:
        return BadRequest(detail=str(self), code=self.code)


class MissingHeader(HeaderError):
    code = "missing_header"

    def __init__(self, header: str) -> None:
        self.header = header
        super().__init__(f"Header missing in request: {header}")


class InvalidHeader(HeaderError):
    def __init__(self, header: str) -> None:
        self.header = header
        super().__init__(f"Malformed header in request: {header}")


def parse_header_datetime(request: Request, header: str) -> datetime.datetime:
    try:
        value = request.headers[header]
    except KeyError:
        raise MissingHeader(header)
    try:
        return datetime.datetime.fromtimestamp(
            parse_http_date(value), tz=datetime.timezone.utc
        )
    except ValueError as e:
        raise InvalidHeader(header) from e


# This can be inlined with a walrus expression but we need to keep support for pre 3.8.
def clean_tags(tags: Iterable[str]) -> Iterable[str]:
    for tag in tags:
        cleaned = tag.strip().strip('"')
        if cleaned:
            yield cleaned


def parse_header_etags(request: Request, header: str) -> FrozenSet[str]:
    try:
        parts = request.headers[header].split(",")
    except KeyError:
        raise MissingHeader(header)
    tags = frozenset(clean_tags(parts))
    if not tags:
        raise InvalidHeader(header)
    return tags



from rest_framework import status
from rest_framework.exceptions import APIException


class PreconditionFailed(APIException):
    status_code = status.HTTP_412_PRECONDITION_FAILED
    default_detail = "An HTTP precondition failed"
    default_code = "precondition_failed"


class BadRequest(APIException):
    status_code = status.HTTP_400_BAD_REQUEST
    default_detail = "Validation failed"
    default_code = "bad_request"



VERSION = (2, 1, 0, "final", 0)


def get_version() -> str:
    """Derives a PEP386-compliant version number from VERSION."""
    assert len(VERSION) == 5
    assert VERSION[3] in ("alpha", "beta", "rc", "final")

    # Now build the two parts of the version number:
    # main = X.Y[.Z]
    # sub = .devN - for pre-alpha releases
    #     | {a|b|c}N - for alpha, beta and rc releases

    parts = 2 if VERSION[2] == 0 else 3
    main = ".".join(str(x) for x in VERSION[:parts])

    sub = ""
    if VERSION[3] != "final":
        mapping = {"alpha": "a", "beta": "b", "rc": "c"}
        sub = mapping[VERSION[3]] + str(VERSION[4])

    return main + sub


__version__ = get_version()



from typing import TYPE_CHECKING, Any, Type, cast

from rest_framework.viewsets import ViewSetMixin

if TYPE_CHECKING:
    from ..mixins import BananasAPI


class BananasBaseRouter:
    def get_default_basename(self, viewset: Type[ViewSetMixin]) -> str:
        return cast(Type["BananasAPI"], viewset).get_admin_meta().basename  # type: ignore[no-any-return]

    def get_schema_view(self) -> Any:
        raise NotImplementedError



from typing import Any, Callable, Dict, List, Optional, Sequence, Tuple, cast

from django.conf import settings
from django.urls.exceptions import NoReverseMatch
from django.utils.translation import gettext as _
from drf_yasg import openapi
from drf_yasg.generators import EndpointEnumerator, OpenAPISchemaGenerator
from drf_yasg.inspectors.view import SwaggerAutoSchema
from drf_yasg.views import get_schema_view
from rest_framework import permissions, viewsets
from rest_framework.authentication import SessionAuthentication
from rest_framework.request import Request
from rest_framework.routers import SimpleRouter
from rest_framework.schemas.coreapi import is_custom_action

from ..versioning import BananasVersioning
from .base import BananasBaseRouter


class BananasEndpointEnumerator(EndpointEnumerator):
    def should_include_endpoint(
        self,
        path: str,
        callback: Callable,
        app_name: str = "",
        namespace: str = "",
        url_name: Optional[str] = None,
    ) -> bool:
        # Fall back to check namespace on the resolver match
        request = self.request
        if (
            not namespace
            and getattr(request, "version", None)
            and getattr(request, "resolver_match", None)
        ):
            namespace = request.resolver_match.namespace or ""
        return cast(
            bool,
            super().should_include_endpoint(
                path, callback, app_name, namespace, url_name
            ),
        )


class BananasOpenAPISchemaGenerator(OpenAPISchemaGenerator):
    endpoint_enumerator_class = BananasEndpointEnumerator

    def get_schema(
        self, request: Optional[Request] = None, public: bool = False
    ) -> Dict[str, Any]:
        schema: Dict[str, Any] = super().get_schema(request, public)
        api_settings = getattr(settings, "ADMIN", {}).get("API", {})
        schema["schemes"] = api_settings.get("SCHEMES", schema["schemes"])
        return schema

    def get_paths(
        self, endpoints: Dict[str, Any], components: Any, request: Request, public: bool
    ) -> Tuple[Any, str]:
        paths, prefix = super().get_paths(endpoints, components, request, public)
        path = request._request.path
        return paths, path[: path.rfind("/")] + prefix


class BananasSwaggerSchema(SwaggerAutoSchema):
    def get_operation_id(self, operation_keys: Sequence[str]) -> str:
        name = ".".join(operation_keys[2:])
        basename: str = self.view.get_admin_meta().basename
        return basename + ":" + name

    def get_summary_and_description(self) -> Tuple[Optional[str], str]:
        """
        Compat: drf-yasg 1.12+
        """
        summary = self.get_summary()
        _, description = super().get_summary_and_description()
        return summary, description

    def get_summary(self) -> str:
        """
        Compat: drf-yasg 1.11
        """
        title = None

        method_name = getattr(self.view, "action", self.method.lower())
        action = getattr(self.view, method_name, None)
        action_kwargs = getattr(action, "kwargs", None)

        if action_kwargs:
            title = action_kwargs.get("name")

        if not title and is_custom_action(self.view.action):
            title = _(self.view.action.replace("_", " ")).capitalize()

        if not title:
            meta = self.view.get_admin_meta()
            if self.view.action in ["retrieve", "update", "partial_update"]:
                title = str(meta.get("verbose_name") or meta.name)
            elif self.view.action == "create":
                title = meta.get("verbose_name")
                if title:
                    title = str(_("Add")) + " " + str(title).lower()
                else:
                    title = meta.name
            elif self.view.action == "list":
                title = str(meta.get("verbose_name_plural") or meta.name)
            else:
                title = str(meta.name)

        return title

    def get_tags(self, operation_keys: Tuple[str, ...]) -> List[str]:
        view = self.view
        meta = self.view.get_admin_meta()
        tags = {f"app:{meta.app_label}"}

        if self.is_navigation():
            tags.add("navigation")

        if issubclass(view.__class__, viewsets.ModelViewSet):
            tags.add("crud")

        view_method = getattr(view, view.action, None)
        if view_method:
            include_tags = set(getattr(view_method, "include_tags", None) or [])
            exclude_tags = set(getattr(view_method, "exclude_tags", None) or [])
            tags |= include_tags
            tags -= exclude_tags

        return [tag for tag in tags if tag not in meta.exclude_tags]

    def is_navigation(self) -> bool:
        if not hasattr(self, "_is_navigation"):
            self._is_navigation = False
            try:
                if self.method == "GET" and (
                    self.view.action == "list" or not hasattr(self.view, "list")
                ):
                    self.view.reverse_action("list")
                    self._is_navigation = True
            except NoReverseMatch:
                pass

        return self._is_navigation


class BananasSimpleRouter(BananasBaseRouter, SimpleRouter):
    def get_schema_view(self) -> Any:
        view = get_schema_view(
            openapi.Info(
                title="Django Bananas Admin API Schema",
                default_version=BananasVersioning.default_version,
                description="API for django-bananas.js",
                # terms_of_service="https://www.google.com/policies/terms/",
                # license=openapi.License(name="BSD License"),
            ),
            # validators=["flex", "ssv"],
            public=False,
            generator_class=BananasOpenAPISchemaGenerator,
            authentication_classes=(SessionAuthentication,),
            permission_classes=(permissions.AllowAny,),
            patterns=self.urls,
        )
        view.versioning_class = BananasVersioning

        return view



from drf_yasg.utils import (
    swagger_auto_schema as schema,
    swagger_serializer_method as schema_serializer_method,
)

from .yasg import (
    BananasSimpleRouter as BananasRouter,
    BananasSwaggerSchema as BananasSchema,
)

__all__ = (
    "schema",
    "schema_serializer_method",
    "BananasRouter",
    "BananasSchema",
)



from django.contrib.auth.models import Permission
from django.core.management.base import BaseCommand, CommandError


class Command(BaseCommand):

    help = "Create admin permissions"

    def handle(self, *args: object, **options: object) -> None:
        if args:  # pragma: no cover
            raise CommandError("Command doesn't accept any arguments")
        return self.handle_noargs(**options)

    def handle_noargs(self, *args: object, **options: object) -> None:
        from django.contrib import admin as django_admin
        from django.contrib.contenttypes.models import ContentType

        from bananas import admin

        django_admin.autodiscover()

        for model, _ in admin.site._registry.items():
            if issubclass(getattr(model, "View", object), admin.AdminView):
                meta = model._meta
                assert isinstance(meta.object_name, str)

                ct, created = ContentType.objects.get_or_create(
                    app_label=meta.app_label, model=meta.object_name.lower()
                )

                if created:
                    print(f"Found new admin view: {ct.name} [{ct.app_label}]")

                for codename, name in model._meta.permissions:
                    p, created = Permission.objects.update_or_create(
                        codename=codename, content_type=ct, defaults={"name": name}
                    )
                    if created:
                        print(f"Created permission: {name}")



from typing import Callable, Optional, Sequence, TypeVar

from django.http.response import HttpResponseBase

T = TypeVar("T", bound=Callable[..., HttpResponseBase])


def tags(
    include: Optional[Sequence[str]] = None, exclude: Optional[Sequence[str]] = None
) -> Callable[[T], T]:
    def decorator(view: T) -> T:
        view.include_tags = include  # type: ignore[attr-defined]
        view.exclude_tags = exclude  # type: ignore[attr-defined]
        return view

    return decorator



import base64
import binascii
import math
import os
import uuid
from itertools import chain
from typing import Any, Dict, Mapping, Optional, Sized

from django.core.exceptions import ValidationError
from django.db import models
from django.utils.translation import gettext_lazy as _
from typing_extensions import Final


class Missing:
    ...


MISSING: Final = Missing()


class ModelDict(Dict[str, Any]):

    _nested: Optional[Dict[str, "ModelDict"]] = None

    def __getattr__(self, item: str) -> Any:
        """
        Try to to get attribute as key item.
        Fallback on prefixed nested keys.
        Finally fallback on real attribute lookup.
        """
        try:
            return self.__getitem__(item)
        except KeyError:
            try:
                return self.__getnested__(item)
            except KeyError:
                return self.__getattribute__(item)

    def __getnested__(self, item: str) -> "ModelDict":
        """
        Find existing items prefixed with given item
        and return a new ModelDict containing matched keys,
        stripped from prefix.

        :param str item: Item prefix key to find
        :return ModelDict:
        """
        # Ensure _nested cache
        if self._nested is None:
            self._nested: Dict[str, ModelDict] = {}

        # Try to get previously accessed/cached nested item
        value = self._nested.get(item, MISSING)

        if not isinstance(value, Missing):
            # Return previously accessed nested item
            return value

        else:
            # Find any keys matching nested prefix
            prefix = item + "__"
            keys = [key for key in self.keys() if key.startswith(prefix)]

            if keys:
                # Construct nested dict of matched keys, stripped from prefix
                n = ModelDict({key[len(item) + 2 :]: self[key] for key in keys})

                # Cache and return
                self._nested[item] = n
                return n

        # Item not a nested key, raise
        raise KeyError(item)

    def expand(self) -> "ModelDict":
        keys = list(self)
        for key in keys:
            field, __, nested_key = key.partition("__")
            if nested_key:
                if field not in keys:
                    nested = self.__getnested__(field)
                    if isinstance(nested, self.__class__):
                        nested = nested.expand()
                    self[field] = nested
                del self[key]
        return ModelDict(self)

    @classmethod
    # Ignore types until no longer work-in-progress.
    def from_model(cls, model, *fields, **named_fields):  # type: ignore[no-untyped-def]
        """
        Work-in-progress constructor,
        consuming fields and values from django model instance.
        """
        d = ModelDict()

        if not (fields or named_fields):
            # Default to all fields
            fields = [  # type: ignore[assignment]
                f.attname for f in model._meta.concrete_fields
            ]

        not_found = object()

        for name, field in chain(zip(fields, fields), named_fields.items()):
            _fields = field.split("__")
            value = model
            for i, _field in enumerate(_fields, start=1):
                # NOTE: we don't want to rely on hasattr here
                previous_value = value
                value = getattr(previous_value, _field, not_found)

                if value is not_found:
                    if _field in dir(previous_value):
                        raise ValueError(
                            "{!r}.{} had an AttributeError exception".format(
                                previous_value, _field
                            )
                        )
                    else:
                        raise AttributeError(
                            "{!r} does not have {!r} attribute".format(
                                previous_value, _field
                            )
                        )

                elif value is None:
                    if name not in named_fields:
                        name = "__".join(_fields[:i])
                    break

            d[name] = value

        return d


class TimeStampedModel(models.Model):
    """
    Provides automatic date_created and date_modified fields.
    """

    date_created = models.DateTimeField(
        blank=True,
        null=True,
        editable=False,
        auto_now_add=True,
        verbose_name=_("date created"),
    )
    date_modified = models.DateTimeField(
        blank=True,
        null=True,
        editable=False,
        auto_now=True,
        verbose_name=_("date modified"),
    )

    class Meta:
        abstract = True

    def save(self, *args: Any, **kwargs: Any) -> None:
        if "update_fields" in kwargs and "date_modified" not in kwargs["update_fields"]:
            update_fields = list(kwargs["update_fields"])
            update_fields.append("date_modified")
            kwargs["update_fields"] = update_fields
        super().save(*args, **kwargs)


class UUIDModel(models.Model):
    """
    Provides auto-generating UUIDField as the primary key for a model.
    """

    id = models.UUIDField(primary_key=True, editable=False, default=uuid.uuid4)

    class Meta:
        abstract = True


class SecretField(models.CharField):
    description = _("Generates and stores a random key.")

    default_error_messages = {
        "random-is-none": _("%(cls)s.get_random_bytes returned None"),
        "random-too-short": _(
            "Too few random bytes received from "
            "get_random_bytes. Number of"
            " bytes=%(num_bytes)s,"
            " min_length=%(min_length)s"
        ),
    }

    def __init__(
        self,
        verbose_name: Optional[str] = None,
        num_bytes: int = 32,
        min_bytes: int = 32,
        auto: bool = True,
        **kwargs: Any,
    ):
        self.num_bytes, self.auto, self.min_length = num_bytes, auto, min_bytes

        field_length = self.get_field_length(self.num_bytes)

        defaults: Mapping[str, Any] = {
            "max_length": field_length,
            **kwargs,
            **(
                {
                    "editable": False,
                    "blank": True,
                }
                if self.auto
                else {}
            ),
        }

        super().__init__(verbose_name, **defaults)

    @staticmethod
    def get_field_length(num_bytes: int) -> int:
        """
        Return the length of hexadecimal byte representation of ``n`` bytes.

        :param num_bytes:
        :return: The field length required to store the byte representation.
        """
        return num_bytes * 2

    def pre_save(self, model_instance: models.Model, add: bool) -> Any:
        if self.auto and add:
            value = self.get_random_str()
            setattr(model_instance, self.attname, value)
            return value
        else:
            return super().pre_save(model_instance, add)

    def get_random_str(self) -> str:
        random = self.get_random_bytes()
        self._check_random_bytes(random)
        return binascii.hexlify(random).decode("utf8")

    def _check_random_bytes(self, random: Optional[Sized]) -> None:
        if random is None:
            raise ValidationError(
                self.error_messages["random-is-none"],
                code="invalid",
                params={"cls": self.__class__.__name__},
            )

        if len(random) < self.min_length:
            raise ValidationError(
                self.error_messages["random-too-short"],
                code="invalid",
                params={"num_bytes": len(random), "min_length": self.min_length},
            )

    def get_random_bytes(self) -> bytes:
        return os.urandom(self.num_bytes)


class URLSecretField(SecretField):
    @staticmethod
    def get_field_length(num_bytes: int) -> int:
        """
        Get the maximum possible length of a base64 encoded bytearray of
        length ``length``.

        :param num_bytes: The length of the bytearray
        :return: The worst case length of the base64 result.
        """
        return math.ceil(num_bytes / 3.0) * 4

    @staticmethod
    def y64_encode(s: bytes) -> bytes:
        """
        Implementation of Y64 non-standard URL-safe base64 variant.

        See http://en.wikipedia.org/wiki/Base64#Variants_summary_table

        :return: base64-encoded result with substituted
        ``{"+", "/", "="} => {".", "_", "-"}``.
        """
        first_pass = base64.urlsafe_b64encode(s)
        return first_pass.translate(bytes.maketrans(b"+/=", b"._-"))

    def get_random_str(self) -> str:
        random = self.get_random_bytes()
        self._check_random_bytes(random)
        return self.y64_encode(random).decode("utf-8")



import logging
from typing import (
    TYPE_CHECKING,
    Any,
    Generic,
    Iterable,
    Iterator,
    List,
    Mapping,
    Optional,
    Type,
    TypeVar,
    Union,
)

from django.db.models import Model
from django.db.models.expressions import Combinable
from django.db.models.query import QuerySet
from typing_extensions import Protocol

from .models import ModelDict

if TYPE_CHECKING:
    from django.db.models.query import _QuerySet

_log = logging.getLogger(__name__)

T = TypeVar("T", bound="QuerySet[Model]")


class ModelDictIterable:
    def __init__(self, queryset: T) -> None:
        self.queryset = queryset
        self.named_fields: Mapping[str, str] = self.queryset._hints.get("_named_fields")  # type: ignore[attr-defined]

    def __iter__(self) -> Iterator[ModelDict]:
        queryset = self.queryset
        query = queryset.query
        compiler = query.get_compiler(queryset.db)

        field_names: List[str] = list(query.values_select)
        extra_names: List[str] = list(query.extra_select)
        annotation_names: List[str] = list(query.annotation_select)

        # Modified super(); rename fields given in queryset.values() kwargs
        names = extra_names + field_names + annotation_names
        if self.named_fields:
            names = self.rename_fields(names)

        for row in compiler.results_iter():
            yield ModelDict(zip(names, row))

    def rename_fields(self, names: Iterable[str]) -> List[str]:
        named_fields = {value: key for key, value in self.named_fields.items()}
        names = [named_fields.get(name, name) for name in names]
        return names


_MT_co = TypeVar("_MT_co", bound=Model, covariant=True)


class IsQuerySet(Protocol[_MT_co]):
    def values(
        self, *fields: Union[str, Combinable], **expressions: Any
    ) -> "_QuerySet[_MT_co, ModelDict]":
        ...


class ModelDictQuerySetMixin:
    def dicts(
        self: IsQuerySet[_MT_co], *fields: str, **named_fields: str
    ) -> "_QuerySet[_MT_co, ModelDict]":
        if named_fields:
            fields += tuple(named_fields.values())

        clone = self.values(*fields)
        clone._iterable_class = ModelDictIterable  # type: ignore[attr-defined]

        # QuerySet._hints is a dict object used by db router
        # to aid deciding which db should get a request. Currently
        # django only supports `instance`, so it's probably
        # fine to set a custom key on this dict as it's a guaranteed
        # way that it'll be returned with the QuerySet instance
        # while leaving the queryset intact
        clone._add_hints(**{"_named_fields": named_fields})  # type: ignore[attr-defined]

        return clone


_MT = TypeVar("_MT", bound=Model)


if TYPE_CHECKING:

    class ModelDictQuerySet(  # type: ignore[misc]
        ModelDictQuerySetMixin,
        QuerySet[_MT],
        IsQuerySet[_MT],
        Generic[_MT],
    ):
        pass

else:

    class ModelDictQuerySet(ModelDictQuerySetMixin, QuerySet):
        pass


class IsManager(Protocol[_MT]):
    model: Type[_MT]
    _db: Optional[str]


class ModelDictManagerMixin:
    def dicts(
        self, *fields: str, **named_fields: str
    ) -> "_QuerySet[_MT_co, ModelDict]":
        # Mypy: `self` types don't add up
        queryset = self.get_queryset()  # type: ignore[misc]
        return queryset.dicts(*fields, **named_fields)

    def get_queryset(self: IsManager[_MT]) -> ModelDictQuerySet:
        return ModelDictQuerySet(self.model, using=self._db)


ExtendedQuerySet = ModelDictQuerySet  # Left for compatibility



from typing import TYPE_CHECKING, Any, Optional, Sequence, Type, TypeVar, cast

from django.db.models import Model
from rest_framework.authentication import BaseAuthentication, SessionAuthentication
from rest_framework.permissions import IsAdminUser
from rest_framework.reverse import reverse
from rest_framework.utils import formatting
from rest_framework.versioning import BaseVersioning
from rest_framework.views import APIView
from rest_framework.viewsets import GenericViewSet, ViewSetMixin

from bananas.models import ModelDict

from .schemas import BananasSchema
from .versioning import BananasVersioning

if TYPE_CHECKING:
    from rest_framework.permissions import _PermissionClass
    from rest_framework.serializers import BaseSerializer

UNDEFINED = object()


class BananasAPI:
    versioning_class: Optional[Type[BaseVersioning]] = BananasVersioning
    authentication_classes: Sequence[Type[BaseAuthentication]] = (
        SessionAuthentication,
    )
    permission_classes: Sequence["_PermissionClass"] = (IsAdminUser,)
    swagger_schema = BananasSchema  # for DRF: schema = BananasSchema()

    _admin_meta: ModelDict
    basename: str

    @classmethod
    def get_admin_meta(cls) -> ModelDict:
        meta: Optional[ModelDict] = getattr(cls, "_admin_meta", None)

        if meta is None:
            # TODO: Get proper app_label, not only root package
            app_label, __, __ = cls.__module__.lower().partition(".")
            name = cls.get_view_name(cls)  # type: ignore[arg-type]

            basename = getattr(cls, "basename", None)
            if basename is None:
                if type(name).__name__ == "__proxy__":
                    # name is lazy, probably gettext, extract basename from class name
                    basename = cls.get_view_name(cls, respect_name=False)  # type: ignore[arg-type]
                else:
                    basename = name
                basename = basename.replace(" ", "_").lower()

            meta = ModelDict(
                app_label=app_label,
                basename=basename,
                name=name,
                exclude_tags=[],
                verbose_name=None,
                verbose_name_plural=name,
            )

            admin = getattr(cls, "Admin", None)
            if admin is not None:
                meta.update(
                    {
                        key: getattr(admin, key)  # type: ignore[misc, call-overload]
                        for key in filter(
                            lambda key: key in meta,  # type: ignore[arg-type]
                            admin.__dict__.keys(),
                        )
                    }
                )

            basename = f"{meta.app_label}.{meta.basename}"
            meta.update({"basename": basename})
            cls._admin_meta = meta

        return meta

    def reverse_action(self, url_name: str, *args: Any, **kwargs: Any) -> str:
        """
        Extended DRF with fallback to requested namespace if request.version is missing
        """
        request = cast(APIView, self).request
        if request and not request.version:
            return reverse(self.get_url_name(url_name), *args, **kwargs)

        return cast(ViewSetMixin, super()).reverse_action(url_name, *args, **kwargs)

    def get_url_name(self, action_url_name: str = "list") -> str:
        """
        Get full namespaced url name to use for reverse()
        """
        url_name = f"{self.basename}-{action_url_name}"

        request = cast(APIView, self).request
        assert request.resolver_match
        namespace = request.resolver_match.namespace
        if namespace:
            url_name = f"{namespace}:{url_name}"

        return url_name

    def get_view_name(self, respect_name: bool = True) -> str:
        """
        Get or generate human readable view name.
        Extended version from DRF to support usage from both class and instance.
        """
        view: Type
        if isinstance(self, type):
            view = self
        else:
            view = self.__class__

        # Name may be set by some Views, such as a ViewSet.
        if respect_name:
            name: Optional[str] = getattr(view, "name", None)
            if name is not None:
                return name

        name = view.__name__
        for view_suffix in ("ViewSet", "View", "API", "Admin"):
            name = formatting.remove_trailing_string(name, view_suffix)
        name = formatting.camelcase_to_spaces(name)

        # Suffix may be set by some Views, such as a ViewSet.
        suffix: Optional[str] = getattr(view, "suffix", None)
        if suffix:
            name += " " + suffix

        return name


_MT_co = TypeVar("_MT_co", bound=Model, covariant=True)


class SchemaSerializerMixin:
    def get_serializer_class(self) -> Type["BaseSerializer[_MT_co]"]:
        serializer_class = cast(GenericViewSet, super()).get_serializer_class()

        action = getattr(self, cast(GenericViewSet, self).action, None)
        schema = getattr(action, "_swagger_auto_schema", None)
        if schema:
            responses = schema.get("responses")
            if responses:
                status_code = sorted(responses.keys())[0]
                if status_code < 300:
                    serializer_class = responses[status_code]

        return serializer_class



from django.urls import include, re_path

from .. import views
from ..router import register, router

register(views.LoginAPI)
register(views.LogoutAPI)
register(views.MeAPI)
register(views.ChangePasswordAPI)
register(views.TranslationAPI)

schema_view = router.get_schema_view()

urlpatterns = [
    re_path(
        r"^schema(?P<format>\.json|\.yaml)$",
        schema_view.without_ui(cache_timeout=0),
        name="schema",
    ),
    re_path(
        r"^swagger$", schema_view.with_ui("swagger", cache_timeout=0), name="swagger"
    ),
    re_path(r"^$", schema_view.with_ui("redoc", cache_timeout=0), name="root"),
    re_path(r"^", include(router.urls)),
]



from types import ModuleType
from typing import Dict, Sequence

from rest_framework.request import Request
from rest_framework.versioning import NamespaceVersioning

from . import v1_0

__versions__ = [v1_0]


class BananasVersioning(NamespaceVersioning):

    default_version: str = v1_0.__version__
    allowed_versions: Sequence[str] = tuple(
        version.__version__ for version in __versions__
    )
    version_map: Dict[str, ModuleType] = {
        version.__version__: version for version in __versions__
    }

    def get_versioned_viewname(self, viewname: str, request: Request) -> str:
        """
        Prefix viewname with full namespace bananas:vX.Y:
        """
        assert request.resolver_match is not None
        namespace = request.resolver_match.namespace
        if namespace:
            viewname = f"{namespace}:{viewname}"

        return viewname



from django.urls import include, re_path

from .versioning import __versions__

apipatterns = [
    re_path(
        rf"^{version.__version__}/",
        include(
            (f"{version.__name__}.urls", "bananas"),
            namespace=version.__version__,
        ),
    )
    for version in __versions__
]


urlpatterns = [re_path(r"^", include((apipatterns, "bananas"), namespace="bananas"))]



import sys
from collections import OrderedDict
from typing import Callable, Dict, List, Optional, Tuple, Union

from django.core.management import BaseCommand
from django.urls import URLPattern, URLResolver, get_resolver


def collect_urls(
    urls: Union[URLResolver, URLPattern, Tuple[str, Callable], None] = None,
    namespace: Optional[str] = None,
    prefix: Optional[list] = None,
) -> List["OrderedDict"]:
    if urls is None:
        urls = get_resolver(urlconf=None)
    prefix = prefix or []
    if isinstance(urls, URLResolver):
        res = []
        pattern = urls.pattern.regex.pattern
        for x in urls.url_patterns:
            res += collect_urls(
                x, namespace=urls.namespace or namespace, prefix=prefix + [pattern]
            )
        return res
    elif isinstance(urls, URLPattern):
        lookup_str = urls.lookup_str
        pattern = urls.pattern.regex.pattern
        return [
            OrderedDict(
                [
                    ("namespace", namespace),
                    ("name", urls.name),
                    ("pattern", prefix + [pattern]),
                    ("lookup_str", lookup_str),
                    ("default_args", dict(urls.default_args or {})),
                ]
            )
        ]
    else:  # pragma: no cover
        raise NotImplementedError(repr(urls))


def show_urls() -> None:
    all_urls = collect_urls()

    max_lengths: Dict[str, int] = {}
    for u in all_urls:
        for k in ["pattern", "default_args"]:
            u[k] = str(u[k])
        for k, v in list(u.items())[:-1]:
            u[k] = v = v or ""
            # Skip app_list due to length (contains all app names)
            if (u["namespace"], u["name"], k) == ("admin", "app_list", "pattern"):
                continue
            max_lengths[k] = max(len(v), max_lengths.get(k, 0))

    for u in sorted(all_urls, key=lambda x: (x["namespace"], x["name"])):
        sys.stdout.write(
            " | ".join(
                ("{:%d}" % max_lengths.get(k, len(v))).format(v) for k, v in u.items()
            )
            + "\n"
        )


class Command(BaseCommand):
    def handle(self, *args: object, **kwargs: object) -> None:
        show_urls()



import abc
import datetime
import operator
from functools import wraps
from typing import (
    Any,
    Callable,
    FrozenSet,
    Generic,
    List,
    NoReturn,
    Optional,
    TypeVar,
    cast,
)

from django.conf import settings
from django.core.exceptions import ImproperlyConfigured
from django.db.models import Model, QuerySet
from drf_yasg import openapi
from drf_yasg.utils import swagger_auto_schema
from rest_framework.mixins import UpdateModelMixin
from rest_framework.request import Request
from rest_framework.response import Response
from rest_framework.serializers import BaseSerializer, ModelSerializer
from rest_framework.viewsets import GenericViewSet
from typing_extensions import Final, Protocol, final

from bananas.admin.api.schemas.yasg import BananasSwaggerSchema
from bananas.models import TimeStampedModel

from . import errors
from .utils import HeaderError, parse_header_datetime, parse_header_etags

__all__ = (
    "Fence",
    "FencedUpdateModelMixin",
    "header_date_parser",
    "parse_date_modified",
    "allow_if_unmodified_since",
    "header_etag_parser",
    "allow_if_match",
)


TokenType = TypeVar("TokenType")
InstanceType = TypeVar("InstanceType")


@final
class Fence(abc.ABC, Generic[InstanceType, TokenType]):
    def __init__(
        self,
        get_token: Callable[[Request], TokenType],
        compare: Callable[[TokenType, TokenType], bool],
        get_version: Callable[[InstanceType], Optional[TokenType]],
        openapi_parameter: openapi.Parameter,
        rejection: Exception = errors.PreconditionFailed(
            "The resource does not fulfill the given preconditions"
        ),
    ) -> None:
        self._get_token: Final = get_token
        self._compare: Final = compare
        self._get_version: Final = get_version
        self._rejection: Final = rejection
        self.openapi_parameter: Final = openapi_parameter

    def check(self, request: Request, instance: InstanceType) -> bool:
        version = self._get_version(instance)
        if version is None:
            # We might want to expose control of this behavior. For if-unmodified-since
            # it makes sense to return true here, but it might not for
            # if-modified-since other conditionals.
            return True
        return self._compare(version, self._get_token(request))

    def reject(self) -> NoReturn:
        raise self._rejection

    def validate(self, request: Request, instance: InstanceType) -> None:
        if not self.check(request, instance):
            self.reject()


class FenceAwareSwaggerAutoSchema(BananasSwaggerSchema):
    update_methods = ("PUT", "PATCH")

    def add_manual_parameters(
        self, parameters: List[openapi.Parameter]
    ) -> List[openapi.Parameter]:
        parameters = super().add_manual_parameters(parameters)
        if (
            isinstance(self.view, FencedUpdateModelMixin)
            and self.method in self.update_methods
        ):
            return parameters + [self.view.fence.openapi_parameter]
        return parameters


_MT_co = TypeVar("_MT_co", bound=Model, covariant=True)


class UsesQuerySet(Protocol[_MT_co]):
    def get_queryset(self) -> "QuerySet[_MT_co]":
        ...


class FencedUpdateModelMixin(UpdateModelMixin, abc.ABC):
    @property
    @abc.abstractmethod
    def fence(self) -> Fence:
        ...

    # django-restframework uses an "advanced self-type" on self in
    # perform_update() which subtly breaks subclassing. We try to remedy this by
    # implementing get_queryset() so that we fulfill the self-type in the
    # overridden method.
    # See https://github.com/typeddjango/djangorestframework-stubs/issues/132
    def get_queryset(self: UsesQuerySet[_MT_co]) -> "QuerySet[_MT_co]":
        return cast(UsesQuerySet[_MT_co], super()).get_queryset()

    def perform_update(self, serializer: BaseSerializer) -> None:
        # mypy's advanced self types don't work with super calls so we use an assertion
        # here instead.
        assert isinstance(self, GenericViewSet)
        assert isinstance(serializer, ModelSerializer)
        self.fence.validate(self.request, serializer.instance)
        super().perform_update(serializer)

    @swagger_auto_schema(auto_schema=FenceAwareSwaggerAutoSchema)
    def update(self, request: Request, *args: Any, **kwargs: Any) -> Response:
        return super().update(request, *args, **kwargs)

    @swagger_auto_schema(auto_schema=FenceAwareSwaggerAutoSchema)
    def partial_update(self, request: Request, *args: Any, **kwargs: Any) -> Response:
        return super().partial_update(request, *args, **kwargs)


def header_date_parser(header: str) -> Callable[[Request], datetime.datetime]:
    def parse(request: Request) -> datetime.datetime:
        try:
            return parse_header_datetime(request, header)
        except HeaderError as e:
            raise e.as_api_error()

    return parse


def parse_date_modified(instance: TimeStampedModel) -> Optional[datetime.datetime]:
    return (
        instance.date_modified.replace(microsecond=0)
        if instance.date_modified
        else None
    )


def allow_if_unmodified_since() -> Fence[TimeStampedModel, datetime.datetime]:
    if not settings.USE_TZ:
        raise ImproperlyConfigured(
            "{} does not support settings.USE_TZ=False.".format(
                allow_if_unmodified_since.__name__
            )
        )
    return Fence(
        get_token=header_date_parser("If-Unmodified-Since"),
        compare=operator.le,
        get_version=parse_date_modified,
        openapi_parameter=openapi.Parameter(
            in_=openapi.IN_HEADER,
            name="If-Unmodified-Since",
            type=openapi.TYPE_STRING,
            required=True,
            description=(
                "Time of last edit of the client's representation of the resource in "
                "RFC7231 format."
            ),
        ),
    )


def header_etag_parser(header: str) -> Callable[[Request], FrozenSet[str]]:
    def parse(request: Request) -> FrozenSet[str]:
        try:
            return parse_header_etags(request, header)
        except HeaderError as e:
            raise e.as_api_error()

    return parse


T = TypeVar("T")


def as_set(
    fn: Callable[[InstanceType], Optional[T]]
) -> Callable[[InstanceType], Optional[FrozenSet[T]]]:
    @wraps(fn)
    def wrapper(instance: InstanceType) -> Optional[FrozenSet[T]]:
        version = fn(instance)
        return None if version is None else frozenset({version})

    return wrapper


def allow_if_match(
    version_getter: Callable[[InstanceType], Optional[str]]
) -> Fence[InstanceType, FrozenSet[str]]:
    return Fence(
        get_token=header_etag_parser("If-Match"),
        compare=operator.le,
        get_version=as_set(version_getter),
        openapi_parameter=openapi.Parameter(
            in_=openapi.IN_HEADER,
            name="If-Match",
            type=openapi.TYPE_STRING,
            required=True,
            description=(
                "Version or list of versions of the clients representation of the "
                "resource."
            ),
        ),
    )



from rest_framework.permissions import BasePermission
from rest_framework.request import Request
from rest_framework.views import APIView


class IsAnonymous(BasePermission):
    """
    Allows access only to non-authenticated users.
    """

    def has_permission(self, request: Request, view: APIView) -> bool:
        return not request.user or request.user.is_anonymous



from typing import TYPE_CHECKING, Type, TypeVar, cast

from rest_framework.viewsets import ViewSetMixin

from .schemas import BananasRouter

if TYPE_CHECKING:
    from .mixins import BananasAPI

__all__ = ["register"]


T = TypeVar("T", bound=ViewSetMixin)


def register(view: Type[T]) -> None:
    """
    Register the API view class in the bananas router.

    :param BananasAPI view:
    """
    meta = cast("BananasAPI", view).get_admin_meta()
    prefix = meta.basename.replace(".", "/")
    router.register(prefix, view, meta.basename)


router = BananasRouter()



"""
--------------------------------------------------------------------------------
 bananas.url - Create Django database configuration from database URIs
--------------------------------------------------------------------------------

Parse database information from a URL, kind of like SQLAlchemy.

================================================================================
 Engines
================================================================================

Currently supported engines are:

==============================  ===========================================
 URI scheme                     Engine
==============================  ===========================================
 pgsql, postgres, postgresql    django.db.backends.postgresql_psycopg2
 mysql                          django.db.backends.mysql
 oracle                         django.db.backends.oracle
 sqlite, sqlite3                django.db.backends.sqlite3
 mysqlgis                       django.contrib.gis.db.backends.mysql
 oraclegis                      django.contrib.gis.db.backends.oracle
 postgis                        django.contrib.gis.db.backends.postgis
 spatialite                     django.contrib.gis.db.backends.spatialite
==============================  ===========================================

You can add your own by running ``register(scheme, module_name)`` before
parsing.
"""
from typing import Any, Dict, List, Mapping, NamedTuple, Optional, Tuple, Union
from urllib.parse import parse_qs, unquote_plus, urlsplit

from typing_extensions import Final


class Alias:
    """
    An alias object used to resolve aliases for engine names.
    """

    def __init__(self, target: str) -> None:
        self.target = target

    def __repr__(self) -> str:
        return f'<Alias to "{self.target}">'


_EngineReference = Union[Alias, str, List[Union[str, Dict[str, str]]]]
ENGINE_MAPPING: Final[Dict[str, _EngineReference]] = {
    "pgsql": Alias("postgresql"),
    "postgres": Alias("postgresql"),
    "postgresql": "django.db.backends.postgresql_psycopg2",
    "mysql": "django.db.backends.mysql",
    "oracle": "django.db.backends.oracle",
    "sqlite": Alias("sqlite3"),
    "sqlite3": "django.db.backends.sqlite3",
    "mysqlgis": "django.contrib.gis.db.backends.mysql",
    "oraclegis": "django.contrib.gis.db.backends.oracle",
    "postgis": "django.contrib.gis.db.backends.postgis",
    "spatialite": "django.contrib.gis.db.backends.spatialite",
}


def register_engine(
    scheme: str,
    module: Union[Alias, str, List[Union[str, Dict[str, str]]]],
) -> None:
    """
    Register a new engine.

    :param scheme: The scheme that should be matched
    :param module:
    :return:
    """
    ENGINE_MAPPING.update({scheme: module})


def resolve(
    cursor: Mapping[str, _EngineReference], key: str
) -> Union[List[Union[str, Dict[str, str]]], str]:
    """
    Get engine or raise exception, resolves Alias-instances to a sibling target.

    :param cursor: The object so search in
    :param key: The key to get
    :return: The object found
    """
    try:
        result = cursor[key]

        # Resolve alias
        if isinstance(result, Alias):
            result = cursor[result.target]

        assert not isinstance(
            result, Alias
        ), "Multiple levels of aliases are not supported"

        return result
    except KeyError:
        raise KeyError("No matches for engine %s" % key)


def get_engine(scheme: str) -> str:
    """
    Perform a lookup in ENGINE_MAPPING using engine_string.

    :param scheme: '+'-separated string Maximum of 2 parts,
    i.e "postgres+psycopg" is OK, "postgres+psycopg2+postgis" is NOT OK.
    :return: Engine string
    """
    path = scheme.split("+")
    first, rest = path[0], path[1:]

    second = rest[0] if rest else None

    engine: Union[str, List[Union[str, Dict[str, str]]], Dict[str, str]]

    engine = resolve(ENGINE_MAPPING, first)

    # If the selected engine does not have a second level.
    if not isinstance(engine, list):
        # If second level engine was expected
        if second:
            raise KeyError("%s has no sub-engines" % first)

        return engine

    try:
        engine, extra = engine
    except ValueError:
        # engine was not a list of length 2
        raise ValueError(
            "django-bananas.url' engine "
            "configuration is invalid: %r" % ENGINE_MAPPING
        )

    assert isinstance(
        extra, Mapping
    ), "The second element in an engine configuration must be a mapping"

    # Get second-level engine
    if second is not None:
        engine = resolve(extra, second)

    # Sanity-check the value before returning
    assert not isinstance(
        engine, (list, dict)
    ), "Only two levels of engines are allowed"
    assert engine, "The returned engine is not truthy"

    return engine


class DatabaseInfo(NamedTuple):
    engine: str
    name: Optional[str]
    schema: Optional[str]
    user: Optional[str]
    password: Optional[str]
    host: Optional[str]
    port: Optional[int]
    params: Dict[str, str]


def parse_path(path: str) -> Tuple[Optional[str], Optional[str]]:
    """
    Get database name and database schema from path.

    :param path: "/"-delimited path, parsed as
     "/<database name>/<database schema>"
    :return: tuple with (database or None, schema or None)
    """
    if path is None:
        raise ValueError("path must be a string")

    parts = path.strip("/").split("/")

    database = unquote_plus(parts[0]) if len(parts) else None
    schema = parts[1] if len(parts) > 1 else None

    return database, schema


def database_conf_from_url(url: str) -> Dict[str, Any]:
    """
    Return a django-style database configuration based on ``url``.

    :param url: Database URL
    :return: Django-style database configuration dict

    Example:
    >>> conf = database_conf_from_url(
    ...     'pgsql://joar:hunter2@5monkeys.se:4242/tweets/tweetschema'
    ...     '?hello=world')
    >>> sorted(conf.items())  # doctest: +NORMALIZE_WHITESPACE
    [('ENGINE', 'django.db.backends.postgresql_psycopg2'),
     ('HOST', '5monkeys.se'),
     ('NAME', 'tweets'),
     ('PARAMS', {'hello': 'world'}),
     ('PASSWORD', 'hunter2'),
     ('PORT', 4242),
     ('SCHEMA', 'tweetschema'),
     ('USER', 'joar')]
    """
    return {key.upper(): val for key, val in parse_database_url(url)._asdict().items()}


def parse_database_url(url: str) -> DatabaseInfo:
    """
    Parse a database URL and return a DatabaseInfo named tuple.

    :param url: Database URL
    :return: DatabaseInfo instance

    Example:
    >>> conf = parse_database_url(
    ...     'pgsql://joar:hunter2@5monkeys.se:4242/tweets/tweetschema'
    ...     '?hello=world')
    >>> conf  # doctest: +NORMALIZE_WHITESPACE
    DatabaseInfo(engine='django.db.backends.postgresql_psycopg2',
                 name='tweets',
                 schema='tweetschema',
                 user='joar',
                 password='hunter2',
                 host='5monkeys.se',
                 port=4242,
                 params={'hello': 'world'})
    """
    if url == "sqlite://:memory:":
        raise Exception(
            'Your url is "sqlite://:memory:", if you want '
            'an sqlite memory database, just use "sqlite://"'
        )
    url_parts = urlsplit(url)

    engine = get_engine(url_parts.scheme)
    database, schema = parse_path(url_parts.path)
    port = url_parts.port
    host = url_parts.hostname
    user = url_parts.username
    password = url_parts.password

    # Take the last element of every parameter list
    params = {key: val.pop() for key, val in parse_qs(url_parts.query).items()}

    return DatabaseInfo(
        engine=engine,
        name=database,
        schema=schema,
        user=user,
        password=password,
        host=host,
        port=port,
        params=params,
    )



from typing import Callable, Iterable, List, Optional, TypeVar

from django.contrib.auth import get_user_model
from django.contrib.auth.models import AbstractUser
from django.contrib.auth.password_validation import password_validators_help_texts
from django.utils.translation import gettext_lazy as _
from rest_framework import serializers

from bananas.admin.api.schemas import schema_serializer_method

User = TypeVar("User", bound=AbstractUser)


class UserSerializer(serializers.ModelSerializer):
    username = serializers.CharField(source="get_username", read_only=True)
    full_name = serializers.SerializerMethodField()
    email = serializers.SerializerMethodField()
    is_superuser = serializers.BooleanField(read_only=True)
    permissions = serializers.SerializerMethodField()
    groups = serializers.SerializerMethodField()

    class Meta:
        model = get_user_model()
        ref_name = "Me"
        fields = (
            "id",
            "username",
            "full_name",
            "email",
            "is_superuser",
            "permissions",
            "groups",
        )
        read_only_fields = fields

    @schema_serializer_method(
        serializer_or_field=serializers.CharField(
            help_text=_("Falls back to username, if not implemented or empty")
        )
    )
    def get_full_name(self, obj: User) -> str:
        getter: Optional[Callable[[], str]] = getattr(obj, "get_full_name", None)
        if getter is not None:
            full_name = getter()

        if not full_name:
            full_name = obj.get_username()

        return full_name

    @schema_serializer_method(serializer_or_field=serializers.CharField())
    def get_email(self, obj: User) -> Optional[str]:
        return getattr(obj, obj.get_email_field_name(), None)

    @schema_serializer_method(
        serializer_or_field=serializers.ListField(
            help_text=_(
                "Permissions that the user has, both through group and user permissions."
            )
        )
    )
    def get_permissions(self, obj: User) -> List[str]:
        return sorted(obj.get_all_permissions())

    @schema_serializer_method(serializer_or_field=serializers.ListField)
    def get_groups(self, obj: User) -> Iterable[str]:
        return obj.groups.order_by("name").values_list("name", flat=True)


class AuthenticationSerializer(serializers.Serializer):
    username = serializers.CharField(label=_("Username"), write_only=True)
    password = serializers.CharField(label=_("Password"), write_only=True)


class PasswordChangeSerializer(serializers.Serializer):
    old_password = serializers.CharField(label=_("Old password"), write_only=True)
    new_password1 = serializers.CharField(
        label=_("New password"),
        help_text="\n".join(password_validators_help_texts()),
        write_only=True,
    )
    new_password2 = serializers.CharField(
        label=_("New password confirmation"), write_only=True
    )



import re
from typing import (
    Any,
    Callable,
    Dict,
    List,
    Optional,
    Tuple,
    Type,
    TypeVar,
    Union,
    cast,
    overload,
)

import django
from django.apps import apps
from django.conf import settings as django_settings
from django.contrib.admin import AdminSite, ModelAdmin
from django.contrib.admin.sites import site as django_admin_site
from django.contrib.auth.decorators import permission_required, user_passes_test
from django.core.handlers.wsgi import WSGIRequest
from django.db.models import Model
from django.http import HttpRequest
from django.http.response import HttpResponse, HttpResponseBase
from django.shortcuts import render
from django.urls import URLPattern, URLResolver, re_path, reverse, reverse_lazy
from django.utils.encoding import force_str
from django.utils.functional import cached_property
from django.utils.safestring import SafeText, mark_safe
from django.utils.translation import gettext_lazy as _
from django.views.generic import View

from ..environment import env

__all__ = ["ModelAdminView", "ViewTool", "AdminView", "register", "site"]


MT = TypeVar("MT", bound=Model)


class ExtendedAdminSite(AdminSite):
    enable_nav_sidebar = False
    default_settings = {
        "INHERIT_REGISTERED_MODELS": env.get_bool(
            "DJANGO_ADMIN_INHERIT_REGISTERED_MODELS", True
        ),
        "SITE_TITLE": env.get("DJANGO_ADMIN_SITE_TITLE", AdminSite.site_title),
        "SITE_HEADER": env.get("DJANGO_ADMIN_SITE_HEADER", "admin"),
        "SITE_VERSION": env.get("DJANGO_ADMIN_SITE_VERSION", django.__version__),
        "INDEX_TITLE": env.get("DJANGO_ADMIN_INDEX_TITLE", AdminSite.index_title),
        "PRIMARY_COLOR": env.get("DJANGO_ADMIN_PRIMARY_COLOR", "#34A77B"),
        "SECONDARY_COLOR": env.get("DJANGO_ADMIN_SECONDARY_COLOR", "#20AA76"),
        "LOGO": env.get("DJANGO_ADMIN_LOGO", "admin/bananas/img/django.svg"),
        "LOGO_ALIGN": env.get("DJANGO_ADMIN_LOGO_ALIGN", "middle"),
        "LOGO_STYLE": env.get("DJANGO_ADMIN_LOGO_STYLE"),
    }
    settings: Dict[str, Any]

    def __init__(self, name: str = "admin") -> None:
        super().__init__(name=name)
        self.settings = dict(self.default_settings)
        self.settings.update(getattr(django_settings, "ADMIN", {}))

        self.site_title = self.settings["SITE_TITLE"]
        self.site_header = self.settings["SITE_HEADER"]
        self.index_title = self.settings["INDEX_TITLE"]

    def each_context(self, request: HttpRequest) -> Dict[str, Any]:
        context = super().each_context(request)
        context.update(settings=self.settings)
        return context

    @property
    def urls(self) -> Tuple[List[Union[URLResolver, URLPattern]], str, str]:
        if self.settings["INHERIT_REGISTERED_MODELS"]:
            for model, admin in list(django_admin_site._registry.items()):
                # django_admin_site.unregister(model)
                self._registry[model] = admin.__class__(model, self)
        return self.get_urls(), "admin", self.name


class ModelAdminView(ModelAdmin):
    @cached_property
    def access_permission(self) -> str:
        meta = self.model._meta
        return "{app_label}.{codename}".format(
            app_label=meta.app_label,
            codename=meta.permissions[0][0],  # First perm codename
        )

    def get_urls(self) -> List[URLPattern]:
        app_label = self.model._meta.app_label
        View: Type[AdminView] = self.model.View
        info = app_label, View.label
        urlpatterns = [
            re_path(
                r"^$",
                self.admin_view(View.as_view(admin=self)),
                name="{}_{}".format(*info),
            ),
            # We add the same url here with _changelist to make sure the
            # admin app index reverse urls to correct view.
            re_path(
                r"^$",
                self.admin_view(View.as_view(admin=self)),
                name="{}_{}_changelist".format(*info),
            ),
        ]
        extra_urls = self.model.View(admin=self).get_urls()
        if extra_urls:
            urlpatterns += extra_urls
        return urlpatterns

    def admin_view(
        self, view: Callable[..., HttpResponseBase], perm: Optional[str] = None
    ) -> Callable[..., HttpResponseBase]:
        if perm is not None:
            perm = self.get_permission(perm)
        else:
            perm = self.access_permission

        admin_login_url = reverse_lazy("admin:login")
        view = user_passes_test(
            lambda u: u.is_active and u.is_staff, login_url=admin_login_url
        )(view)
        view = permission_required(perm, login_url=admin_login_url)(view)
        return view

    def get_permission(self, perm: str) -> str:
        if "." not in perm:
            perm = f"{self.model._meta.app_label}.{perm}"
        return perm

    def has_module_permission(self, request: HttpRequest) -> bool:
        return request.user.has_perm(self.access_permission)

    def has_change_permission(
        self, request: HttpRequest, obj: Optional[MT] = None
    ) -> bool:
        return request.user.has_perm(self.access_permission)

    # TODO: Remove obj?
    def has_add_permission(
        self, request: HttpRequest, obj: Optional[MT] = None
    ) -> bool:
        return False

    def has_delete_permission(
        self, request: HttpRequest, obj: Optional[MT] = None
    ) -> bool:
        return False

    def get_context(self, request: WSGIRequest, **extra: Any) -> Dict[str, Any]:
        opts = self.model._meta
        context = self.admin_site.each_context(request)
        context.update(
            {
                "app_label": opts.app_label,
                "model_name": force_str(opts.verbose_name_plural),
                "title": force_str(opts.verbose_name_plural),
                "cl": {"opts": opts},  # change_list.html requirement
                "opts": opts,  # change_form.html requirement
                "media": self.media,
            }
        )
        context.update(extra or {})
        return context


# Call without parenthesis: @register
@overload
def register(
    view: Type["AdminView"],
    *,
    admin_site: Optional[AdminSite] = None,
    admin_class: Type[ModelAdmin] = ModelAdminView,
) -> Type["AdminView"]:
    ...


# Call with parenthesis: @register()
@overload
def register(
    view: None = ...,
    *,
    admin_site: Optional[AdminSite] = None,
    admin_class: Type[ModelAdmin] = ModelAdminView,
) -> Callable[[Type["AdminView"]], Type["AdminView"]]:
    ...


def register(
    view: Optional[Type["AdminView"]] = None,
    *,
    admin_site: Optional[AdminSite] = None,
    admin_class: Type[ModelAdmin] = ModelAdminView,
) -> Union[Type["AdminView"], Callable[[Type["AdminView"]], Type["AdminView"]]]:
    """
    Register a generic class based view wrapped with ModelAdmin and fake model

    :param view: The AdminView to register.
    :param admin_site: The AdminSite to register the view on.
        Defaults to bananas.admin.ExtendedAdminSite.
    :param admin_class: The ModelAdmin class to use for eg. permissions.
        Defaults to bananas.admin.ModelAdminView.

    Example:

    @register  # Or with args @register(admin_class=MyModelAdminSubclass)
    class MyAdminView(bananas.admin.AdminView):
        def get(self, request):
            return self.render('template.html', {})

    # Also possible:
    register(MyAdminView, admin_class=MyModelAdminSublass)

    """
    if not admin_site:
        admin_site = site

    def wrapped(inner_view: Type["AdminView"]) -> Type["AdminView"]:
        module = inner_view.__module__
        match = re.search(r"\.?(\w+)\.admin", module)
        assert match is not None
        app_label = match.group(1)
        app_config = apps.get_app_config(app_label)

        label = getattr(inner_view, "label", None)
        if not label:
            label = re.sub("(Admin)|(View)", "", inner_view.__name__).lower()
        inner_view.label = label

        model_name = label.capitalize()
        verbose_name = getattr(inner_view, "verbose_name", model_name)
        inner_view.verbose_name = verbose_name

        access_perm_codename = "can_access_" + model_name.lower()
        access_perm_name = _("Can access {verbose_name}").format(
            verbose_name=verbose_name
        )
        # The first permission here is expected to be
        # the general access permission.
        permissions = tuple(
            [(access_perm_codename, access_perm_name)]
            + list(getattr(inner_view, "permissions", []))
        )

        model = type(
            model_name,
            (Model,),
            {
                "__module__": module + ".__models__",  # Fake
                "View": inner_view,
                "app_config": app_config,
                "Meta": type(
                    "Meta",
                    (object,),
                    {
                        "managed": False,
                        "abstract": True,
                        "app_label": app_config.label,
                        "verbose_name": verbose_name,
                        "verbose_name_plural": verbose_name,
                        "permissions": permissions,
                    },
                ),
            },
        )

        assert admin_site is not None
        admin_site._registry[model] = admin_class(model, admin_site)
        return inner_view

    if view is None:  # Used as a decorator
        return wrapped

    return wrapped(view)


class ViewTool:
    def __init__(
        self, text: str, link: str, perm: Optional[str] = None, **attrs: Any
    ) -> None:
        self.text = text
        self._link = link
        self.perm = perm

        html_class = attrs.pop("html_class", None)
        if html_class is not None:
            attrs.setdefault("class", html_class)
        self._attrs = attrs

    @property
    def attrs(self) -> SafeText:
        return mark_safe(" ".join(f"{k}={v}" for k, v in self._attrs.items()))

    @property
    def link(self) -> str:
        if "/" not in self._link:
            return reverse(self._link)
        return self._link


class AdminView(View):
    tools: Optional[List[Union[Tuple[str, str], Tuple[str, str, str], ViewTool]]] = None
    action: Optional[str] = None
    admin: Optional[ModelAdminView] = None

    label: str
    verbose_name: str
    request: WSGIRequest

    def dispatch(
        self, request: HttpRequest, *args: Any, **kwargs: Any
    ) -> HttpResponseBase:
        # Try to fetch set action first.
        # This should be the view name for custom views
        action = self.action
        if action is None:
            return super().dispatch(request, *args, **kwargs)

        handler: Callable[..., HttpResponseBase] = getattr(self, action)
        return handler(request, *args, **kwargs)

    def get_urls(self) -> List[URLPattern]:
        """Should return a list of urls
        Views should be wrapped in `self.admin_view` if the view isn't
        supposed to be accessible for non admin users.
        Omitting this can cause threading issues.

        Example:

            return [
                url(r'^custom/$',
                    self.admin_view(self.custom_view))
            ]
        """
        return []

    def get_tools(
        self,
    ) -> Optional[List[Union[Tuple[str, str], Tuple[str, str, str], ViewTool]]]:
        # Override point, self.request is available.
        return self.tools

    def get_view_tools(self) -> List[ViewTool]:
        tools = []
        all_tools = self.get_tools()
        if all_tools:
            for tool in all_tools:
                if isinstance(tool, (list, tuple)):
                    perm = None
                    # Mypy doesn't change type on a len(...) call
                    # See: https://github.com/python/mypy/issues/1178
                    if len(tool) == 3:
                        tool, perm = cast(Tuple[str, str, str], tool)[:-1], tool[-1]
                    text, link = cast(Tuple[str, str], tool)
                    tool = ViewTool(text, link, perm=perm)
                else:
                    # Assume ViewTool
                    perm = tool.perm
                if perm and not self.has_permission(perm):
                    continue

                tools.append(tool)

        return tools

    def admin_view(
        self,
        view: Callable[..., HttpResponseBase],
        perm: Optional[str] = None,
        **initkwargs: Any,
    ) -> Callable[..., HttpResponseBase]:
        assert self.admin is not None
        view = self.__class__.as_view(
            action=view.__name__, admin=self.admin, **initkwargs
        )
        return self.admin.admin_view(view, perm=perm)

    def get_permission(self, perm: str) -> str:
        assert self.admin is not None
        return self.admin.get_permission(perm)

    def has_permission(self, perm: str) -> bool:
        perm = self.get_permission(perm)
        return self.request.user.has_perm(perm)

    def has_access(self) -> bool:
        assert self.admin is not None
        return self.has_permission(self.admin.access_permission)

    def get_context(self, **extra: Any) -> Dict[str, Any]:
        assert self.admin is not None
        return self.admin.get_context(
            self.request, view_tools=self.get_view_tools(), **extra
        )

    def render(
        self, template: str, context: Optional[Dict[str, Any]] = None
    ) -> HttpResponse:
        extra = context or {}
        return render(self.request, template, self.get_context(**extra))


site = ExtendedAdminSite()



import logging
from functools import partial
from os import environ
from typing import (
    Any,
    Callable,
    Dict,
    Generic,
    Iterable,
    List,
    Optional,
    Set,
    Tuple,
    Type,
    TypeVar,
    Union,
    cast,
)

from django.conf import global_settings
from typing_extensions import Final, Protocol, overload

__all__ = ["env", "parse_bool", "parse_int", "parse_tuple", "parse_list", "parse_set"]


log = logging.getLogger(__name__)


class Undefined:
    ...


UNDEFINED: Final = Undefined()

UNSUPPORTED_ENV_SETTINGS = (
    "ADMINS",
    "MANAGERS",
    "LANGUAGES",
    "DISALLOWED_USER_AGENTS",
    "IGNORABLE_404_URLS",
    "TEMPLATES",
)

SETTINGS_TYPES: Dict[str, type] = {
    "LANGUAGE_COOKIE_AGE": int,
    "EMAIL_TIMEOUT": int,
    "FILE_UPLOAD_PERMISSIONS": int,
    "FILE_UPLOAD_DIRECTORY_PERMISSIONS": int,
}


def parse_str(value: str) -> str:
    """
    Clean string.

    :param str value: Original str.
    :return: str: Cleaned str.
    """
    return value.strip()


def parse_bool(value: str) -> bool:
    """
    Parse string to bool.

    :param str value: String value to parse as bool
    :return bool:
    """
    boolean = parse_str(value).capitalize()

    if boolean in ("True", "Yes", "On", "1"):
        return True
    elif boolean in ("False", "No", "Off", "0"):
        return False
    else:
        raise ValueError(f'Unable to parse boolean value "{value}"')


def parse_int(value: str) -> int:
    """
    Parse numeric string to int. Supports oct formatted string.

    :param str value: String value to parse as int
    :return int:
    """
    value = parse_str(value=value)
    if value != "0" and value.startswith("0"):
        return int(value.lstrip("0o"), 8)
    else:
        return int(value)


Q = TypeVar("Q", covariant=True)


class _Instantiable(Protocol[Q]):
    def __init__(self, value: Iterable[Q]) -> None:
        ...


class _InstantiableIterable(Iterable[Q], _Instantiable[Q], Generic[Q]):
    ...


T = TypeVar("T", bound=_InstantiableIterable)


def parse_iterable(typ: Type[T], value: str) -> T:
    """
    Parse comma separated string into typed iterable.

    :param typ: Type to cast iterable to
    :param value: String value to parse as iterable
    :return: Given type
    """
    return typ(parse_str(v) for v in value.split(","))


parse_tuple = cast(Callable[[str], Tuple[str, ...]], partial(parse_iterable, tuple))
parse_list = cast(Callable[[str], List[str]], partial(parse_iterable, list))
parse_set = cast(Callable[[str], Set[str]], partial(parse_iterable, set))


Builtin = Union[str, bool, int]
B = TypeVar("B", bound=Builtin)
P = TypeVar("P", bound=Union[Builtin, tuple, list, set])


@overload
def get_parser(typ: Type[B]) -> Callable[[str], B]:
    ...


@overload
def get_parser(typ: Type[tuple]) -> Callable[[str], Tuple[str, ...]]:
    ...


@overload
def get_parser(typ: Type[list]) -> Callable[[str], List[str]]:
    ...


@overload
def get_parser(typ: Type[set]) -> Callable[[str], Set[str]]:
    ...


def get_parser(typ: Type[P]) -> Callable[[str], P]:
    """
    Return appropriate parser for given type.

    :param typ: Type to get parser for.
    :return function: Parser
    """
    try:
        return cast(
            Callable[[str], P],
            {
                str: parse_str,
                bool: parse_bool,
                int: parse_int,
                tuple: parse_tuple,
                list: parse_list,
                set: parse_set,
            }[typ],
        )
    except KeyError:
        raise NotImplementedError("Unsupported setting type: %r", typ)


def get_settings() -> Dict[str, Any]:
    """
    Get and parse prefixed django settings from env.

    TODO: Implement support for complex settings
        DATABASES = {}
        CACHES = {}
        INSTALLED_APPS -> EXCLUDE_APPS ?

    :return dict:
    """
    settings = {}
    prefix = environ.get("DJANGO_SETTINGS_PREFIX", "DJANGO_")
    parse: Callable[[str], object]

    for key, value in environ.items():
        _, _, key = key.partition(prefix)
        if key:
            if key in UNSUPPORTED_ENV_SETTINGS:
                raise ValueError(
                    'Django setting "{}" can not be '
                    "configured through environment.".format(key)
                )

            default_value = getattr(global_settings, key, UNDEFINED)

            if default_value is not UNDEFINED:
                if default_value is None and key in SETTINGS_TYPES.keys():
                    # Handle typed django settings defaulting to None
                    parse = get_parser(SETTINGS_TYPES[key])
                else:
                    # Determine parser by django setting type
                    parse = get_parser(type(default_value))

                value = parse(value)

            settings[key] = value

    return settings


S = TypeVar("S")
U = TypeVar("U")


class EnvironWrapper:
    """
    Wrapper around os environ with type conversion support.
    """

    def __delitem__(self, key: str) -> None:
        environ.__delitem__(key)

    def __getitem__(self, key: str) -> str:
        return environ.__getitem__(key)

    def __setitem__(self, key: str, value: str) -> None:
        environ.__setitem__(key, value)

    def __contains__(self, key: object) -> bool:
        return environ.__contains__(key)

    def __getattr__(self, item: str) -> object:
        return getattr(environ, item)

    get: Callable[..., str]

    @overload
    def parse(self, parser: Callable[[str], S], key: str, default: None) -> Optional[S]:
        ...

    @overload
    def parse(self, parser: Callable[[str], S], key: str, default: S) -> S:
        ...

    def parse(
        self, parser: Callable[[str], S], key: str, default: Optional[S] = None
    ) -> Optional[S]:
        value: Union[str, Undefined] = environ.get(key, UNDEFINED)
        if isinstance(value, Undefined):
            return default
        try:
            return parser(value)
        except ValueError:
            log.warning(
                "Unable to parse environment variable {key}={value}".format(
                    key=key, value=value
                )
            )
            return default

    @overload
    def get_bool(self, key: str, default: U) -> Union[bool, U]:
        ...

    @overload
    def get_bool(self, key: str, default: None = None) -> Optional[bool]:
        ...

    def get_bool(self, key: str, default: object = None) -> object:
        return self.parse(parse_bool, key, default=default)

    @overload
    def get_int(self, key: str, default: U) -> Union[int, U]:
        ...

    @overload
    def get_int(self, key: str, default: None = None) -> Optional[int]:
        ...

    def get_int(self, key: str, default: object = None) -> object:
        return self.parse(parse_int, key, default=default)

    @overload
    def get_tuple(self, key: str, default: U) -> Union[Tuple[str, ...], U]:
        ...

    @overload
    def get_tuple(self, key: str, default: None = None) -> Optional[Tuple[str, ...]]:
        ...

    def get_tuple(self, key: str, default: object = None) -> object:
        return self.parse(parse_tuple, key, default=default)

    @overload
    def get_list(self, key: str, default: U) -> Union[List[str], U]:
        ...

    @overload
    def get_list(self, key: str, default: None = None) -> Optional[List[str]]:
        ...

    def get_list(self, key: str, default: object = None) -> object:
        return self.parse(parse_list, key, default=default)

    @overload
    def get_set(self, key: str, default: U) -> Union[Set[str], U]:
        ...

    @overload
    def get_set(self, key: str, default: None = None) -> Optional[Set[str]]:
        ...

    def get_set(self, key: str, default: object = None) -> object:
        return self.parse(parse_set, key, default=default)


env = EnvironWrapper()



from typing import Any, Dict

from django.views.i18n import JavaScriptCatalog


class RawTranslationCatalog(JavaScriptCatalog):

    domain = "django"

    def render_to_response(self, context: Dict[str, Any], **response_kwargs: Any) -> Dict[str, Any]:  # type: ignore[override]
        return context



from django.contrib.auth import (
    login as auth_login,
    logout as auth_logout,
    update_session_auth_hash,
)
from django.contrib.auth.base_user import AbstractBaseUser
from django.contrib.auth.forms import AuthenticationForm, PasswordChangeForm
from django.utils.translation import gettext_lazy as _
from rest_framework import serializers, status, viewsets
from rest_framework.permissions import AllowAny
from rest_framework.request import Request
from rest_framework.response import Response

from .i18n import RawTranslationCatalog
from .mixins import BananasAPI
from .permissions import IsAnonymous
from .schemas import schema
from .serializers import (
    AuthenticationSerializer,
    PasswordChangeSerializer,
    UserSerializer,
)

UNDEFINED = object()


class BananasAdminAPI(BananasAPI, viewsets.GenericViewSet):
    pass


class LoginAPI(BananasAdminAPI):

    name = _("Log in")
    basename = "login"
    permission_classes = (IsAnonymous,)
    serializer_class = AuthenticationSerializer  # Placeholder for schema

    class Admin:
        verbose_name_plural = None

    @schema(responses={200: UserSerializer})
    def create(self, request: Request) -> Response:
        """
        Log in django staff user
        """
        # TODO: Decorate api with sensitive post parameters as Django admin do?
        # from django.utils.decorators import method_decorator
        # from django.views.decorators.debug import sensitive_post_parameters
        # sensitive_post_parameters_m = method_decorator(sensitive_post_parameters())

        login_form = AuthenticationForm(request, data=request.data)

        if not login_form.is_valid():
            raise serializers.ValidationError(login_form.errors)

        auth_login(request, login_form.get_user())

        serializer = UserSerializer(request.user)
        return Response(serializer.data, status=status.HTTP_200_OK)


class LogoutAPI(BananasAPI, viewsets.ViewSet):

    name = _("Log out")
    basename = "logout"

    class Admin:
        verbose_name_plural = None

    @schema(responses={204: ""})
    def create(self, request: Request) -> Response:
        """
        Log out django staff user
        """
        auth_logout(request)
        return Response(status=status.HTTP_204_NO_CONTENT)


class MeAPI(BananasAdminAPI):

    serializer_class = UserSerializer

    class Admin:
        exclude_tags = ["navigation"]

    @schema(responses={200: UserSerializer})
    def list(self, request: Request) -> Response:
        """
        Retrieve logged in user info
        """
        serializer = self.get_serializer(request.user)
        return Response(serializer.data, status=status.HTTP_200_OK)


class ChangePasswordAPI(BananasAdminAPI):

    name = _("Change password")
    basename = "change_password"
    serializer_class = PasswordChangeSerializer  # Placeholder for schema

    class Admin:
        verbose_name_plural = None

    @schema(responses={204: ""})
    def create(self, request: Request) -> Response:
        """
        Change password for logged in django staff user
        """
        # TODO: Decorate api with sensitive post parameters as Django admin do?

        assert isinstance(request.user, AbstractBaseUser)
        password_form = PasswordChangeForm(request.user, data=request.data)

        if not password_form.is_valid():
            raise serializers.ValidationError(password_form.errors)

        password_form.save()
        update_session_auth_hash(request, password_form.user)

        return Response(status=status.HTTP_204_NO_CONTENT)


class TranslationAPI(BananasAdminAPI):

    name = _("Translation catalog")
    basename = "i18n"
    permission_classes = (AllowAny,)

    class Admin:
        exclude_tags = ["navigation"]

    @schema(responses={200: ""})
    def list(self, request: Request) -> Response:
        """
        Retrieve the translation catalog.
        """
        return Response(RawTranslationCatalog().get(request))
